#!/usr/bin/python2.5
# -*- coding: utf-8 -*-

#  PycaWM
#  config.py
#  Copyright (c) 2007 Vincent Rasneur, Anaël Verrier

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 only.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import os
import sys

from Xlib import X

from pycawm import Client
from pycawm import ClientUnderPointer
from pycawm import MoveUnderPointer
from pycawm import PycaWM
from pycawm.button import (CloseButton, MiniIconButton, ShadeButton,
                           MaximizeButton, IconifyButton)
from pycawm.hookmanager import add_post_hook, add_pre_hook
from pycawm.menu import (Menu, FocusesMenu, IconifiedClientsMenu,
                         ClientsSwitchMenu)
from pycawm.plugins import MiniIconBox
from pycawm.plugins import MovePos
from pycawm.plugins import Pager
from pycawm.plugins import RemoteREPL
from pycawm.plugins import StickyRootEdges


# 1- Create the window manager

# The optional argument is the display name.
# If there is no argument, the wm will look for the $DISPLAY environment
# variable. If $DISPLAY is not set, the wm will use ":0.0" as the display name.
pika = PycaWM()

# Save the path to the wm, we will need it to create a restart function
pika_file = os.path.normpath(os.path.join(os.getcwd(), sys.argv[0]))

# 2- Create the key bindings

# Key bindings are set by the add_key method from the PycaWM or Client
# (you can use the remove_key method to remove a key binding)
# A key binding is composed of two arguments:
#   * the keyboard shortcut (modifiers + key).
#   * a function to call when the keys are pressed.

# PycaWM.add_key adds a key binding to the root window.
# This is an instance method.
# The function to call when a key is pressed takes the wm as its only argument.

pika.add_key('C-M-Escape', lambda wm: wm.run_command('xkill'))
pika.add_key('C-M-r', lambda wm: wm.switch_wm('/usr/bin/python2.5', pika_file))
pika.add_key('C-M-q', lambda wm: wm.quit())
pika.add_key('M-Tab', lambda wm: ClientsSwitchMenu(wm, 'M-Tab'))

# Client.add_key adds a key binding to all the clients (i.e. managed windows).
# This is a class method.
# The function to call when a key is pressed takes a client as its only
# argument.

Client.add_key('C-F4', Client.close)
Client.add_key('C-F5', lambda client: client.unmaximize() \
                   if client.states.maximized \
                   else client.fullscreen())
Client.add_key('C-F6', lambda client: client.unmaximize() \
                   if client.states.maximized \
                   else client.take_all_free_space())
Client.add_key('C-F7', lambda client: client.unmaximize() \
                   if client.states.maximized \
                   else client.take_all_horizontal_space())
Client.add_key('C-F8', lambda client: client.unmaximize() \
                   if client.states.maximized \
                   else client.take_all_vertical_space())

# 3- Create the client buttons

# Both Client.add_left_button and Client.add_right_button take a Button object
# as their only argument.

# add the left buttons...
Client.add_left_button(CloseButton('close.png'))
Client.add_left_button(MiniIconButton('plop.png'))

# ... and the right buttons
Client.add_right_button(CloseButton('close.png'))
Client.add_right_button(MaximizeButton('maximize.png', 'unmaximize.png'))
Client.add_right_button(ShadeButton('shade.png', 'unshade.png'))
Client.add_right_button(IconifyButton('iconify.png'))

# 4- Add some hooks

# We want to show some menus when we click on the root window
# So we will hook the wm button press event method.
# The hook function launch_app_on_click will be called each time the wm
# processes a button press event.

# each menu will be placed where the mouse pointer is.
def launch_app_on_click(self, event):
    # did we click on the root window?
    if event.window == self.root:
        # left button?
        if event.detail == 1:
            Menu(self, x=event.root_x, y=event.root_y,
                 elements=(('Root Menu', Menu.separator),
                           ('', Menu.separator),
                           ('xterm', Menu.shell, 'xterm'),
                           ('rox', Menu.shell, 'rox'),
                           ('emacs', Menu.shell, 'emacs'),
                           ('links2', Menu.shell, 'links2', '-g'),
                           ('Gimp', Menu.shell, 'gimp'),
                           ('', Menu.separator),
                           ('liquidwar', Menu.shell, 'liquidwar'),
                           ('frozen-bubble', Menu.shell, 'frozen-bubble'),
                           ('', Menu.separator),
                           ('Move Window', Menu.python,
                            lambda: MoveUnderPointer(self)),
                           ('Close Window', Menu.python,
                            lambda: ClientUnderPointer(self, Client.close)),
                           ('', Menu.separator),
                           ('Restart PycaWM', Menu.python,
                            lambda: self.switch_wm(pika_file)),
                           ('Quit PycaWM', Menu.python, self.quit)))
        # middle button?
        elif event.detail == 2:
            FocusesMenu(self, x=event.root_x, y=event.root_y)
        # right button?
        elif event.detail == 3:
            IconifiedClientsMenu(self, x=event.root_x, y=event.root_y)

# add the hook to the wm at the proper place.
add_post_hook(pika.button_press_reply, launch_app_on_click)

# 4- Set more stuff

# PycaWM has a special internal method to automatically set a wallpaper.
# It just takes the wallpaper path as an argument.
pika.set_wallpaper('wallpaper.png')

# Set the default focus mode.
# A focus mode is just a class with a bunch of hooks to the wm methods.
# The PycaWM.set_focus method takes the class name as an argument.
# Other modes are: ClickToFocus, FocusFollowsMouse, ClickToSloppyFocus.
pika.set_focus('SloppyFocus')
# Set the preferred placements mode.
# A placement mode is a class with a place_client method.
# The PycaWM.set_placements method takes a list of class names as an argument.
# The placements modes are tried successively until the client has been
# correctly placed. If no placement modes worked, the wm will put the
# client at (0, 0).
pika.set_placements(['SmartPlacement', 'CascadePlacement'])

# 5- Add some plugins

# PycaWM can be easily extended by plugins.
# To add a plugin, call the PycaWM.add_plugin method.
# This method takes 2 arguments:
#   * the plugin class.
#   * the arguments we want to give to the plugin when the wm instantiates it.

# Put a pager at (0, 0)
pika.add_plugin(Pager, dict(x=0, y=0))
# Put a miniiconbox at (5, 100). There can be at most 3 iconified clients
# by line. When no miniicon is found, put the image at "plop.png" instead.
pika.add_plugin(MiniIconBox, dict(x=5, y=108,
                                  columns=3,
                                  image_default='plop.png'))
# When we move a client, show its new position.
pika.add_plugin(MovePos)
# Add sticky edges to the root window. The default threshold to stick a window
# onto the border is 20 pixels.
pika.add_plugin(StickyRootEdges)
# Add a REPL. We will be able to inspect and modify the window manager
# at runtime.
# To connect to the REPL, use "pycarepl.py" (in the "utils" directory).
pika.add_plugin(RemoteREPL)

# 6- Last things to do...

# Manage all the existing clients connected to the wm display.
pika.manage_existing_windows()
# Run the wm event loop.
pika.event_loop()
