# -*- coding: utf-8 -*-

#  PycaWM
#  hookmanager.py
#  Copyright (c) 2007 Vincent Rasneur, Anaël Verrier

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 only.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys

if sys.version_info < (2, 5):
    raise ImportError('hookmanager module needs at least Python 2.5!')

from types import FunctionType
from functools import wraps

# dictionaries containing list of hooks
# the key is the hooked function
pre_hooks = dict()
post_hooks = dict()
overriding_hooks = dict()

# dictionaries containing list of hooks, too
# the key is a tuple: (hooked function, instance needed to run the hooks)
instance_pre_hooks = dict()
instance_post_hooks = dict()
instance_overriding_hooks = dict()

def execute_hook(hooks, key, args, kwargs):
    """Helper function for the run_hooks decorator.
    Try to execute the hooks for a given function or (function, instance)."""
    ret_args = None
    if key in hooks:
        # if a hook removes itself, bad things can happen to the list
        # so, copy it
        hooks_key = (hooks[key])[:]
        for hook in hooks_key:
            try:
                ret_args = hook(*args, **kwargs)
            except TypeError:
                pass
    return ret_args

def execute_instance_hook(hooks, fun, args, kwargs):
    """Helper function for the run_hooks decorator.
    Try to execute the hooks for a given (function, instance)."""
    inst = args[0] if args else None
    execute_hook(hooks, (fun, inst), args, kwargs)

def run_hooks(fun):
    """Decorator to make a function hookable."""
    @wraps(fun)
    def wrapper(*args, **kwargs):
        inst = args[0] if args else None
        overrided = False
        execute_hook(pre_hooks, fun, args, kwargs)
        execute_instance_hook(instance_pre_hooks, fun, args, kwargs)
        if fun in overriding_hooks and overriding_hooks[fun]:
            fun_args = execute_hook(overriding_hooks, fun, args, kwargs)
            overrided = True
        if ((fun, inst) in instance_overriding_hooks and
            instance_overriding_hooks[(fun, inst)]):
            fun_args = execute_instance_hook(instance_overriding_hooks,
                                             fun, args, kwargs)
            overrided = True
        if not overrided:
            fun_args = fun(*args, **kwargs)
        execute_hook(post_hooks, fun, args, kwargs)
        execute_instance_hook(instance_post_hooks, fun, args, kwargs)
        return fun_args
    wrapper._wrapped_fun = fun
    return wrapper

def classmethod_hooks(fun):
    """Shortcut for defining a hookable classmethod."""
    return classmethod(run_hooks(fun))

def staticmethod_hooks(fun):
    """Shortcut for defining a hookable staticmethod."""
    return staticmethod(run_hooks(fun))

def add_hook(hooks, fun, hook):
    """Generic function for adding a hook in a predefined hooks dictionary.
    The given hook will be executed each time the function is called."""
    try:
        wrapped_fun = fun._wrapped_fun
        hooks.setdefault(wrapped_fun, list()).insert(0, hook)
    except AttributeError:
        print 'no wrapped fun?!'

def add_instance_hook(hooks, fun, hook):
    """Generic function for adding an instance hook in a predefined
    hooks dictionary.
    The given hook will be executed each time the function is called when
    the function instance the same as the one stored in the hooks key."""
    if not hasattr(fun, 'im_self'):
        print 'the hooked function is not an instance method!'
        return

    inst = fun.im_self
    try:
        wrapped_fun = fun._wrapped_fun
        hooks.setdefault((wrapped_fun, inst), list()).insert(0, hook)
    except AttributeError:
        print 'no wrapped fun?!'

def add_pre_hook(fun, hook):
    """Add a pre-hook to the given function.
    The given hook will be executed before the function."""
    add_hook(pre_hooks, fun, hook)

def add_post_hook(fun, hook):
    """Add a post-hook to the given function.
    The given hook will be executed after the function."""
    add_hook(post_hooks, fun, hook)

def add_overriding_hook(fun, hook):
    """Add an overriding-hook to the given function.
    The given hook will be executed instead of the function.

    If more than one overriding hook are added, they are all executed
    in the order they have been added."""
    add_hook(overriding_hooks, fun, hook)

def add_instance_pre_hook(fun, hook):
    """Add an instance pre-hook to the given function.
    The given hook will be executed before the function."""
    add_instance_hook(instance_pre_hooks, fun, hook)

def add_instance_post_hook(fun, hook):
    """Add an instance post-hook to the given function.
    The given hook will be executed after the function."""
    add_instance_hook(instance_post_hooks, fun, hook)

def add_instance_overriding_hook(fun, hook):
    """Add an instance overriding-hook to the given function.
    The given hook will be executed instead of the function.

    If more than one overriding hook are added, they are all executed
    in the order they have been added."""
    add_instance_hook(instance_overriding_hooks, fun, hook)

def remove_hook(hooks, fun, hook):
    """Generic function for removing a hook in a predefined hooks dictionary."""
    try:
        wrapped_fun = fun._wrapped_fun
        if wrapped_fun in hooks:
            hooks[wrapped_fun].remove(hook)
            # if there is no more hooks, delete the key
            if not hooks[wrapped_fun]:
                del hooks[wrapped_fun]
    except ValueError:
        print 'hook removing failed!'

def remove_instance_hook(hooks, fun, hook):
    """Generic function for removing an instance hook in a predefined
    hooks dictionary."""
    if not hasattr(fun, 'im_self'):
        print 'the hooked function is not an instance method!'
        return

    inst = fun.im_self
    try:
        key = (fun._wrapped_fun, inst)
        if key in hooks:
            hooks[key].remove(hook)
            # if there is no more hooks, delete the key
            if not hooks[key]:
                del hooks[key]
    except ValueError:
        print 'hook removing failed!'

def remove_pre_hook(fun, hook):
    """Remove a pre-hook from the given function."""
    remove_hook(pre_hooks, fun, hook)

def remove_post_hook(fun, hook):
    """Remove a post-hook from the given function."""
    remove_hook(post_hooks, fun, hook)

def remove_overriding_hook(fun, hook):
    """Remove an overriding-hook from the given function."""
    remove_hook(overriding_hooks, fun, hook)

def remove_instance_pre_hook(fun, hook):
    """Remove an instance pre-hook from the given function."""
    remove_instance_hook(instance_pre_hooks, fun, hook)

def remove_instance_post_hook(fun, hook):
    """Remove an instance post-hook from the given function."""
    remove_instance_hook(instance_post_hooks, fun, hook)

def remove_instance_overriding_hook(fun, hook):
    """Remove an instance overriding-hook from the given function."""
    remove_instance_hook(instance_overriding_hooks, fun, hook)

def make_disabling_hook(hooked_fun, nb=1):
    """Make a temporary overriding hook that will do nothing for the number
    of times indicated."""
    cont = [nb]
    def countdown(*args, **kwargs):
        cont[0] -= 1
        if cont[0] <= 0:
            remove_overriding_hook(hooked_fun, countdown)
    return countdown

def disable_function(hooked_fun, nb=1):
    """Shortcut to temporarily disable a function for nb times."""
    add_overriding_hook(hooked_fun, make_disabling_hook(hooked_fun, nb))

def make_instance_disabling_hook(hooked_fun, nb=1):
    """Make a temporary overriding hook that will do nothing for the number
    of times indicated.

    Warning: if the hooked fun is an instance method, the hook will only
    decrement if called with the same instance as the hooked fun."""
    cont = [nb]
    def countdown(*args, **kwargs):
        # not necessary, but make the check:
        # instance method and same self as hooked fun?
        if not (args and hasattr(hooked_fun, 'im_self') and
                args[0] == hooked_fun.im_self):
            remove_instance_overriding_hook(hooked_fun, countdown)
            return

        cont[0] -= 1
        if cont[0] <= 0:
            remove_instance_overriding_hook(hooked_fun, countdown)
    return countdown

def disable_method_for_instance(hooked_meth, nb=1):
    """Shortcut to temporarily disable a method only for a given instance
    for nb times."""
    add_instance_overriding_hook(hooked_meth,
                                 make_instance_disabling_hook(hooked_meth, nb))

class MetaHookable(type):
    """ Metaclass to make all the method from an object automatically
    hookable."""
    # TODO: does it work with multiple inherance for classmethods and
    # staticmethods?
    def __new__(cls, name, bases, classdict):
        for (attr, meth) in classdict.iteritems():
            if not hasattr(meth, '_wrapped_fun'):
                if type(meth) is FunctionType:
                    classdict[attr] = run_hooks(meth)
                elif type(meth) is classmethod:
                    classdict[attr] = classmethod(run_hooks(
                            meth.__get__(None, name).im_func))
                elif type(meth) is staticmethod:
                    classdict[attr] = staticmethod(run_hooks(
                            meth.__get__(None, name)))
        return type.__new__(cls, name, bases, classdict)
